from infi.projector.plugins import CommandPlugin
from infi.projector.helper import assertions
from logging import getLogger

logger = getLogger(__name__)

USAGE = """
Usage:
    projector version release <version> [--no-fetch] (--no-upload | [--distributions=DISTRIBUTIONS] [--pypi-servers=PYPI_SERVERS]) [--no-push-changes] [--keep-leftovers]
    projector version upload <version> [--distributions=DISTRIBUTIONS] [--pypi-servers=PYPI_SERVERS]

Options:
    version release                 Release a new version, including registering and uploading to pypi
    version upload                  Upload an exisiting version to pypi
    <version>                       x.y.z, or the keywords: trivial, minor, major ('release' only); current, latest ('upload' only)
    --no-upload                     Do not upload the package as part of the release process
    --no-fetch                      Do not fetch origin before releasing
    --distributions=DISTRIBUTIONS   Distributions to build [default: sdist,bdist_egg]
    --pypi-servers=PYPI             PyPI server for publishing [default: pypi,]
    --no-push-changes               Do no push release commits and tags to origin
    --keep-leftovers                If something fails during release, don't
"""

class VersionPlugin(CommandPlugin):
    def get_docopt_string(self):
        return USAGE

    def get_command_name(self):
        return 'version'

    @assertions.requires_built_repository
    def parse_commandline_arguments(self, arguments):
        assertions.assert_on_branch("develop")
        assertions.assert_no_uncommitted_changes()
        methods = [self.release, self.upload]
        [method] = [method for method in methods
                    if arguments.get(method.__name__)]
        self.arguments = arguments
        method()

    def replace_version_tag(self):
        """find the next major/minor/trivial version number if applicable"""
        version_tag = self.arguments.get('<version>')
        special_keywords = ['current', 'latest']
        if version_tag in special_keywords:
            logger.error("releasing version '{}' is disallowed. Did you mean 'version upload'?".format(version_tag))
            raise SystemExit(1)
        placeholders = dict(major=0, minor=1, trivial=2)
        placeholder = placeholders.get(version_tag)
        if placeholder is None:
            return version_tag
        current_version = self.get_git_describe().lstrip('v')
        version_numbers = current_version.split('-')[0].split('.')
        version_numbers = [int(item) for item in version_numbers]
        version_numbers = version_numbers[:placeholder + 1]
        while len(version_numbers) < 3:
            version_numbers.append(0)
        version_numbers[placeholder] += 1
        return '.'.join([str(item) for item in version_numbers[:2 if placeholder < 2 else 3]])

    def fetch_origin(self):
        from gitpy import LocalRepository
        from gitpy.exceptions import GitCommandFailedException
        from os import curdir
        repository = LocalRepository(curdir)
        try:
            repository.fetch()
        except (TypeError, GitCommandFailedException), error:
            logger.error("Failed to fetch origin: {}".format(getattr(error, 'msg', error.message)))
            logger.info("Either fix this or run with --no-fetch")
            raise SystemExit(1)

    def release(self):
        from infi.projector.helper import assertions
        from infi.projector.helper.utils import release_version_with_git_flow, git_checkout
        version_tag = self.replace_version_tag()
        if not self.arguments.get('--no-fetch', False):
            self.fetch_origin()
        assertions.assert_version_tag_for_release(version_tag)
        assertions.assert_develop_branch_on_top_of_master()
        assertions.assert_develop_and_master_not_behind_origin()
        version_tag_without_v = version_tag.lstrip('v')
        version_tag_with_v = 'v{}'.format(version_tag_without_v)
        release_version_with_git_flow(version_tag_with_v, self.arguments.get("--keep-leftovers", False))
        self.arguments['<version>'] = version_tag
        push_changes = not self.arguments.get("--no-push-changes", False)
        if push_changes:
            self.push_commits_and_tags()
        if self.arguments.get('--no-upload', False):
            git_checkout("develop")
        else:
            self.upload()

    def upload(self):
        from infi.projector.helper.assertions import assert_version_tag_for_upload
        from infi.projector.helper.utils import git_checkout, get_latest_version
        version_tag = self.arguments['<version>']
        assert_version_tag_for_upload(version_tag)
        if version_tag == 'current':
            version_to_upload = 'HEAD'
        elif version_tag == 'latest':
            version_to_upload = get_latest_version()
        else:
            version_tag_without_v = version_tag.lstrip('v')
            version_to_upload = 'v{}'.format(version_tag_without_v)
        self.build_and_upload_distributions(version_to_upload)
        git_checkout("develop")

    def get_repository(self):
        from gitpy import LocalRepository
        from os import curdir
        return LocalRepository(curdir)

    def push_commits_and_tags(self):
        repository = self.get_repository()
        logger.debug("Pushing changes to origin")
        repository._executeGitCommandAssertSuccess("git push --all")
        repository._executeGitCommandAssertSuccess("git push --tags")

    def get_git_describe(self):
        return self.get_repository()._executeGitCommand("git describe --tags").stdout.read().splitlines()[0]

    def get_current_version_from_git_describe(self):
        returned = self.get_git_describe()
        all_tags = set(tag.name for tag in self.get_repository().getTags())
        if returned not in all_tags:
            returned = "{}.post{}.{}".format(*returned.rsplit("-", 2))
        return returned

    def build_and_upload_distributions(self, version_tag_with_v):
        from infi.projector.helper.utils import execute_with_buildout, git_checkout
        from infi.projector.plugins.builtins.devenv import DevEnvPlugin
        for distribution in self.arguments.get("--distributions").split(','):
            for pypi in self.arguments.get("--pypi-servers").split(','):
                git_checkout(version_tag_with_v)
                DevEnvPlugin().create_setup_py()
                setup_cmd = "setup . register -r {pypi} {distribution} upload -r {pypi}"
                setup_cmd = setup_cmd.format(pypi=pypi, distribution=distribution)
                execute_with_buildout(setup_cmd, env=dict(LC_ALL="C"))



from infi.pyutils.contexts import contextmanager
from infi.projector.plugins import CommandPlugin
from logging import getLogger

logger = getLogger(__name__)

USAGE = """
Usage:
    projector repository init [--mkdir] <project_name> <origin> <short_description> <long_description>
    projector repository clone <origin>
    projector repository skeleton update [--remove-deprecated-files] [--commit-changes]
    projector repository sync <remote-user> <remote-host> [<remote-path>] [--watch] [--verbose]

Options:
    repository init                 Create a new project/git repository
    repository clone                Clone an exisiting project/git repository
    repository skeleton update      Update skeleton-related files (e.g bootstrap.py)
    repository sync                 sync this repository with a remote target
    <project_name>                  The name of the project in python-module-style (object)
    <origin>                        Remote repository url
    <short_description>             A one-line description
    <long_description>              A multi-line description
    <remote-path>                   if missing, assuming target is at the default installation directory
    --mkdir                         Init the repository in a new directory instead of the current directory
    --remove-deprecated-files       remove files that were in use in previous versions of projector but are no longer necessary
    --watch                         watch for changes
"""


def get_package_namespace(name):
    namespaces = []
    for item in name.split('.')[:-1]:
        namespaces.append('.'.join([namespaces[-1], item]) if namespaces else item)
    return namespaces


def generate_package_code():
    from uuid import uuid1
    return '{' + str(uuid1()) + '}'


def indent(text):
    return '\n'.join(['\t{}'.format(line) for line in text.splitlines()]).strip()


def get(original, section, key, default=None):
    from ConfigParser import NoOptionError
    try:
        return original.get(section, key)
    except NoOptionError:
        return default


class RepositoryPlugin(CommandPlugin):
    def get_docopt_string(self):
        return USAGE

    def get_command_name(self):
        return 'repository'

    def parse_commandline_arguments(self, arguments):
        methods = [self.init, self.clone, self.skeleton, self.sync]
        [method] = [method for method in methods
                    if arguments.get(method.__name__)]
        self.arguments = arguments
        method()

    def get_project_name(self):
        return self.arguments.get("<project_name>")

    @contextmanager
    def _create_subdir_if_necessary(self):
        from infi.projector.helper.utils import chdir
        from os.path import exists, isdir, sep
        from os import makedirs, name
        if not self.arguments.get('--mkdir'):
            yield
            return
        dirname = self.arguments.get('<project_name>') or self.arguments.get('<origin>')
        if name == 'nt':
            dirname = dirname.replace(sep, '/')
        dirname = (dirname if not dirname.endswith('.git') else dirname[0:-4]).split('/')[-1]
        if exists(dirname) and isdir(dirname):
            logger.debug("{} already exists".format(dirname))
            raise SystemExit(1)
        makedirs(dirname)
        with chdir(dirname):
            yield

    def _exit_if_dotgit_exists(self):
        from os.path import exists
        if exists('.git'):
            logger.error("This directory is already a git repository")
            raise SystemExit(1)

    def git_init(self):
        from os.path import curdir
        from gitpy.repository import LocalRepository
        repository = LocalRepository(curdir)
        repository.init()
        repository.addRemote("origin", self.arguments.get('<origin>'))

    def gitflow_init(self):
        from gitflow.core import GitFlow
        gitflow = GitFlow()
        gitflow.init(force_defaults=True)

    def release_initial_version(self):
        from infi.projector.helper.utils import release_version_with_git_flow
        release_version_with_git_flow("v0")

    def add_initial_files(self):
        from os.path import basename
        from shutil import copy
        from .skeleton import get_files
        for src, dst in [(filepath, basename(filepath)) for filepath in get_files()]:
            copy(src, dst)

    def set_buildout_config(self):
        from infi.projector.helper.utils import open_buildout_configfile
        project_name = self.get_project_name()
        with open_buildout_configfile(write_on_exit=True) as buildout:
            buildout.set('project', 'name', project_name)
            buildout.set('project', 'namespace_packages', get_package_namespace(project_name))
            buildout.set('project', 'version_file',
                             '/'.join(['src'] + project_name.split('.') + ['__version__.py']))
            buildout.set('project', 'description', self.arguments.get("<short_description>"))
            buildout.set('project', 'long_description', indent(self.arguments.get("<long_description>")))
            buildout.set('project', 'upgrade_code', generate_package_code())
            buildout.set('project', 'product_name', project_name)

    def get_package_directories(self):
        from os.path import sep
        name = self.get_project_name()
        return [item.replace('.', sep) for item in get_package_namespace(name)] + [name.replace('.', sep)]

    def generate_src(self):
        from os import mkdir
        from os.path import join
        file_content = """__import__("pkg_resources").declare_namespace(__name__)\n"""
        mkdir('src')
        for dirname in self.get_package_directories():
            mkdir(join('src', dirname))
            with open(join('src', dirname, '__init__.py'), 'w') as file:
                file.write(file_content)

    def append_to_gitignore(self):
        project_name = self.get_project_name()
        with open('.gitignore', 'a') as fd:
            fd.write('\n' + '/'.join(['src'] + project_name.split('.') + ['__version__.py']) + '\n')
            fd.write("bootstrap.py\n")

    def safe_append_to_gitignore(self, entry):
        with open('.gitignore') as fd:
            entries = [line.strip() for line in fd.readlines()]
        if entry in entries:
            return
        with open('.gitignore', 'a') as fd:
            fd.write(entry + "\n")

    def commit_all(self):
        from os import curdir
        from gitpy import LocalRepository
        repository = LocalRepository(curdir)
        repository.addAll()
        repository.commit("added all project files")

    def git_checkout_develop(self):
        from os import curdir
        from gitpy import LocalRepository
        repository = LocalRepository(curdir)
        repository.checkout("develop")

    def init(self):
        with self._create_subdir_if_necessary():
            self._exit_if_dotgit_exists()
            self.git_init()
            self.gitflow_init()
            self.release_initial_version()
            self.git_checkout_develop()
            self.add_initial_files()
            self.set_buildout_config()
            self.generate_src()
            self.append_to_gitignore()
            self.commit_all()

    def git_clone(self):
        from os import curdir
        from gitpy import LocalRepository
        repository = LocalRepository(curdir)
        origin = self.arguments.get("<origin>")
        logger.debug("Cloning {}".format(origin))
        repository.clone(origin)

    def origin_has_develop_branch(self):
        from os import curdir
        from gitpy import LocalRepository
        from gitpy.exceptions import NonexistentRefException
        repository = LocalRepository(curdir)
        try:
            repository.getRemoteByName("origin").getBranchByName("develop")
            return True
        except NonexistentRefException:
            return False

    def clone(self):
        self.arguments['--mkdir'] = True
        with self._create_subdir_if_necessary():
            self.git_clone()
            if self.origin_has_develop_branch():
                self.git_checkout_develop()
                self.gitflow_init()

    def overwrite_update_files(self):
        from os.path import basename
        from .skeleton import get_files_to_update
        from shutil import copy
        from os import curdir
        from gitpy import LocalRepository
        repository = LocalRepository(curdir)
        for src, dst in [(filepath, basename(filepath)) for filepath in get_files_to_update()]:
            copy(src, dst)
            logger.info("overwriting {}".format(dst))
            if self.arguments.get("--commit-changes", False):
                repository.add(dst)

    def remove_deprecated_files(self):
        from os import curdir, path, remove
        from gitpy import LocalRepository
        repository = LocalRepository(curdir)
        for filename in [filename for filename in ["buildout-git.cfg",
                                                   "buildout-version.cfg",
                                                   "buildout-pack.cfg",
                                                   "buildout-dist.cfg",
                                                   "buildout.in"]
                         if path.exists(filename)]:
            logger.info("removing {}".format(filename))
            if self.arguments.get("--commit-changes", False):
                repository.delete(filename, force=True)
            else:
                remove(filename)

    def skeleton(self):
        ATTRIBURES_BY_SECTION = {'project': ['name', 'namespace_packages', 'install_requires', 'version_file',
                                             'description', 'long_description', 'console_scripts', 'upgrade_code',
                                             'package_data']}
        DEFAULTS = {'namespace_packages': [], 'install_requires': [], 'console_scripts': [], 'package_data': []}
        from infi.projector.helper.utils import open_buildout_configfile
        if not self.arguments.get("update"):
            logger.error("Not implemented")
            raise SystemExit(1)
        logger.info("Starting skeleton update")
        with open_buildout_configfile() as original:
            backup = {}
            logger.info("Backing up buildout sections")
            for section in ATTRIBURES_BY_SECTION.keys():
                backup[section] = {key: get(original, section, key, DEFAULTS.get(key, None))
                                   for key in ATTRIBURES_BY_SECTION[section]}
            if self.arguments.get("--remove-deprecated-files", False):
                logger.info("Removing deprecated files")
                self.remove_deprecated_files()
            logger.info("Writing skeleton files")
            self.overwrite_update_files()
            self.safe_append_to_gitignore("bootstrap.py")
            with open_buildout_configfile(write_on_exit=True) as update:
                logger.info("Writing buildout.cfg")
                for section, attributes in backup.items():
                    for key, value in attributes.items():
                        update.set(section, key, value)

        if self.arguments.get("--commit-changes", False):
            logger.info("Committing changes")
            from gitpy import LocalRepository
            from os import curdir
            repository = LocalRepository(curdir)
            message = "updated project files from skeleton"
            repository.commit(message, commitAll=True)

    def _get_default_remote_path(self):
        from infi.projector.helper.utils import open_buildout_configfile
        is_windows = self.arguments.get("<remote-user>") == "Administrator"
        basedir = "/cygdrive/c/Program Files/" if is_windows else "/opt"
        with open_buildout_configfile() as buildout:
            get = buildout.get
            if is_windows:
                return "/".join([basedir, get("project", "company"), get("project", "product_name")])
            else:
                return "/".join([basedir, get("project", "company").lower(),
                                 get("project", "product_name").replace(' ', '-').replace('_', '-').lower()])

    def sync(self):
        from infi.pysync import main
        args = ["--python"]
        if self.arguments.get("--watch"):
            args.extend(["--watch"])
        if self.arguments.get("--verbose"):
            args.extend(["--verbose"])
        patterns = [".cache", ".git", ".gitignore", ".installed.cfg", ".projector", "MANIFEST.in",
                    "bin", "bootstrap.py", "develop-eggs", "eggs", "parts", "dist", "devlocal", "data", "setup.py",
                    "src/*egg-info", "src/**/__version__.py"]
        if not self.arguments.get("<remote-path>"):
            patterns.extend(["buildout.cfg", "setup.in"])
        args.extend(["--skip-source={}".format(item) for item in patterns])
        args.extend(["--skip-target={}".format(item) for item in patterns])
        default_remote_path = self._get_default_remote_path()
        args.extend(["{}@{}:{}".format(self.arguments.get("<remote-user>"), self.arguments.get("<remote-host>"),
                     self.arguments.get("<remote-path>") or default_remote_path)])
        logger.info("pysync {}".format(" ".join(args)))
        return main(args)



__version__ = "0.6.15"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "arnony@infinidat.com"
__git_branch__ = '(detached from v0.6.15)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '5339eb78dce37b619d7d1f0e26c5b4947c8bfc5b'
__git_head_subject__ = 'Finished release v0.6.15.'
__git_head_message__ = ''
__git_dirty_diff__ = ''


from contextlib import contextmanager
from infi.projector.plugins import CommandPlugin
from infi.projector.helper import assertions, utils
from logging import getLogger

logger = getLogger(__name__)

USAGE = """
Usage:
    projector devenv build [--clean] [--force-bootstrap] [--no-submodules] [--no-setup-py] [--no-scripts] [--no-readline] [--use-isolated-python] [[--newest] | [--offline] | [--prefer-final]]
    projector devenv relocate ([--absolute] | [--relative]) [--commit-changes]
    projector devenv pack

Options:
    devenv build            use this command to generate setup.py and the console scripts
    devenv relocate         use this command to switch from relative and absolute paths in the console scripts
    devenv pack             create a package, e.g. deb/rpm/msi
    --clean                 clean build-related files and directories before building
    --force-bootstrap       run bootstrap.py even if the buildout script already exists
    --no-submodules         do not clone git sub-modules defined in buildout.cfg
    --no-scripts            do not install the dependent packages, nor create the console scripts. just create setup.py
    --no-readline           do not install [py]readline support (where applicable)
    --use-isolated-python   do not use global system python in console scripts, use Infinidat's isolated python builds
    --newest                always check for new package version on PyPI
    --offline               install packages only from download cache (no internet connection)
    --absolute              change the paths in the development environment to absolute paths
    --relative              change the paths in the development environment to relative paths
    --prefer-final          don't install development versions of dependencies, prefer their latest final versions.
"""

class DevEnvPlugin(CommandPlugin):
    def get_docopt_string(self):
        return USAGE

    def get_command_name(self):
        return 'devenv'

    @assertions.requires_repository
    def parse_commandline_arguments(self, arguments):
        methods = [self.build, self.relocate, self.pack]
        [method] = [method for method in methods
                    if arguments.get(method.__name__)]
        self.arguments = arguments
        method()

    def create_cache_directories(self):
        from os import makedirs
        from os.path import join, exists
        with utils.open_buildout_configfile() as buildout:
            cachedir = buildout.get("buildout", "download-cache")
        cache_dist = join(cachedir, "dist")
        if not exists(cache_dist):
            makedirs(cache_dist)

    def _get_pypi_index_url(self):
        from ConfigParser import ConfigParser, NoOptionError, NoSectionError
        from os import path
        pydistutils = ConfigParser()
        pydistutils.read([path.expanduser(path.join("~", basename)) for basename in ['.pydistutils.cfg', 'pydistutils.cfg']])
        try:
            return pydistutils.get("easy_install", "index-url").strip("/")
        except (NoSectionError, NoOptionError):
            return "https://pypi.python.org/simple"

    def _get_bootstrap_command(self):
        from os.path import exists
        from os import environ
        if not exists("bootstrap.py"):
            logger.error("bootsrap.py does not exist")
            raise SystemExit(1)

        cmd = "bootstrap.py"
        additional_optional_args = {"PROJECTOR_BOOTSTRAP_DOWNLOAD_BASE": "--download-base",
                                    "PROJECTOR_BOOTSTRAP_SETUP_SOURCE": "--setup-source",
                                    "PROJECTOR_BOOTSTRAP_INDEX_URL": "--index-url",
                                    }
        for key, cmd_option in additional_optional_args.items():
            option_value = environ.get(key, None)
            if option_value:
                cmd += ' {}={}'.format(cmd_option, option_value)
        if not environ.get("PROJECTOR_BOOTSTRAP_INDEX_URL", None):
            # we want to extract the index-url from pydistutils.cfg
            cmd += ' --index-url={}'.format(self._get_pypi_index_url())
        # in case dependencies are frozen, we need to use the frozen version of setuptools
        with utils.open_buildout_configfile() as buildout:
            if buildout.has_option("versions", "setuptools"):
                cmd += ' --setuptools-version={}'.format(buildout.get("versions", "setuptools"))
        return cmd

    def bootstrap_if_necessary(self):
        from os.path import join
        from pkg_resources import resource_filename
        from infi.projector.plugins.builtins.repository import skeleton
        buildout_executable_exists = assertions.is_executable_exists(join("bin", "buildout"))
        bootstrap_py = resource_filename(skeleton.__name__, "bootstrap.py")
        with open("bootstrap.py", "w") as dst:
            with open(bootstrap_py) as src:
                dst.write(src.read())
        if not buildout_executable_exists or self.arguments.get("--force-bootstrap", False) or self.arguments.get("--newest", False):
            utils.execute_with_python(self._get_bootstrap_command())

    def install_sections_by_recipe(self, recipe):
        with utils.open_buildout_configfile() as buildout:
            sections_to_install = [section for section in buildout.sections()
                                   if buildout.has_option(section, "recipe") and
                                      buildout.get(section, "recipe") == recipe]
        if sections_to_install:
            utils.execute_with_buildout("install {}".format(' '.join(sections_to_install)))

    def submodule_update(self):
        with utils.buildout_parameters_context(['buildout:develop=']):
            self.install_sections_by_recipe("zerokspot.recipe.git")

    def create_setup_py(self):
        with utils.buildout_parameters_context(['buildout:develop=']):
            self.install_sections_by_recipe("infi.recipe.template.version")

    def get_isolated_python_section_name(self):
        with utils.open_buildout_configfile() as buildout:
            sections = [section for section in buildout.sections()
                        if buildout.has_option(section, "recipe") and
                        buildout.get(section, "recipe").startswith("infi.recipe.python")
                        and not buildout.get(section, "recipe").endswith(":pack")]
        return sections[0]

    def create_scripts(self):
        additional_options = ["buildout:prefer-final=true"] if self.arguments.get("--prefer-final") else []
        with utils.buildout_parameters_context(additional_options):
            self.install_sections_by_recipe("infi.recipe.console_scripts")
            self.install_sections_by_recipe("infi.recipe.console_scripts:gui_scripts")

    def _remove_files_of_type_recursively(self, root_path, file_type):
        import os
        file_type = file_type if file_type.startswith(".") else "." + file_type
        [[os.remove(os.path.join(p, f)) for f in fs if f.endswith(file_type)] for p, ds, fs in os.walk(root_path)]

    def clean_build(self):
        from os.path import exists
        from os import remove
        from shutil import rmtree
        directories_to_clean = ['bin', 'eggs', 'develop-eggs']
        files_to_clean = ['setup.py']
        [remove(filename) for filename in files_to_clean if exists(filename)]
        [rmtree(dirname) for dirname in directories_to_clean if exists(dirname)]
        self._remove_files_of_type_recursively("src", "pyc")

    @contextmanager
    def buildout_newest_or_offline_context(self):
        parameters = []
        if self.arguments.get('--newest'):
            parameters.append('-n')
        if self.arguments.get('--offline'):
            parameters.append('-o')
        with utils.buildout_parameters_context(parameters):
            yield

    def get_readline_module(self):
        from platform import system
        modules = {"Darwin": 'readline',
                   "Windows": 'pyreadline'}
        return modules.get(system())

    def is_module_installed(self, module):
        from infi.execute import execute_assert_success, ExecutionError
        try:
            execute_assert_success("bin/python -c import {}".format(module).split())
        except (OSError, ExecutionError):  # pragma: no cover
            return False
        return True

    def install_readline(self):
        from infi.execute import execute_assert_success, ExecutionError
        module = self.get_readline_module()
        if not module or self.is_module_installed(module):  # pragma: no cover
            return
        try:
            execute_assert_success("bin/easy_install {}".format(module).split())
        except (OSError, ExecutionError):  # pragma: no cover
            logger.warn("easy_install script not generated (perhaps not a requirement). "
                        "Not installing readline support")

    def _remove_setuptools_egg_link(self):
        # HOSTDEV-1130
        # https://bugs.launchpad.net/zc.buildout/+bug/1210996
        import os
        from ConfigParser import NoOptionError, NoSectionError
        with utils.open_buildout_configfile() as buildout:
            try:
                develop_eggs_dir = buildout.get("buildout", "develop-eggs-directory")
            except (NoSectionError, NoOptionError):
                develop_eggs_dir = "develop-eggs"
            setuptools_egg_link = os.path.join(develop_eggs_dir, "setuptools.egg-link")
            if os.path.exists(setuptools_egg_link):
                os.remove(setuptools_egg_link)

    def install_isolated_python_if_necessary(self):
        if not self.arguments.get("--use-isolated-python", False):
            return
        self._remove_setuptools_egg_link()
        if not assertions.is_isolated_python_exists() or self.arguments.get("--newest", False):
            with utils.buildout_parameters_context(['buildout:develop=']):
                utils.execute_with_buildout("install {}".format(self.get_isolated_python_section_name()))
            self.arguments["--force-bootstrap"] = True
            utils.execute_with_isolated_python(self._get_bootstrap_command())

    def build(self):
        if self.arguments.get("--clean", False):
            self.clean_build()
        elif self.arguments.get("--newest", False):
            self._remove_files_of_type_recursively("src", "pyc")
        self.create_cache_directories()
        self.bootstrap_if_necessary()
        with self.buildout_newest_or_offline_context():
            self.install_isolated_python_if_necessary()
            if not self.arguments.get("--no-submodules", False):
                self.submodule_update()
            if not self.arguments.get("--no-setup-py", False):
                self.create_setup_py()
            if not self.arguments.get("--no-scripts", False):
                self.create_scripts()
                if not self.arguments.get("--no-readline", False):
                    self.install_readline()

    def relocate(self):
        relative_paths = self.arguments.get("--relative", False)
        with utils.open_buildout_configfile(write_on_exit=True) as buildout:
            buildout.set("buildout", "relative-paths", 'true' if relative_paths else 'false')
        if self.arguments.get("--commit-changes", False):
            commit_message = "Changing shebang to {} paths".format("relative" if relative_paths else "absolute")
            utils.commit_changes_to_buildout(commit_message)
        logger.info("Configuration changed. Run `projector devenv build [--use-isolated-python]`.")

    def pack(self):
        assertions.assert_isolated_python_exists()
        self.install_sections_by_recipe("infi.recipe.application_packager")



from infi.pyutils.lazy import cached_method
from logging import getLogger

logger = getLogger(__name__)

COMMAND_PLUGIN_ENTRY_POINT = "projector_command_plugins"

class CommandPlugin(object):  # pragma: no cover
    def get_docopt_string(self):
        raise NotImplementedError()

    def get_command_name(self):
        raise NotImplementedError()

    def parse_commandline_arguments(self, arguments):
        raise NotImplementedError()

class PluginRepository(object):
    @cached_method
    def get_all_plugins(self):
        return [plugin_class() for plugin_class in self.get_all_plugin_classes()
                if self.validate_plugin(plugin_class)]

    def validate_plugin(self, plugin_class):
        if not issubclass(plugin_class, CommandPlugin):
            logger.error("Plugin {} is not a subclasss of CommandPlugin".format(plugin_class.__name__))
            return False
        methods = ["get_docopt_string", "get_command_name", "parse_commandline_arguments"]
        for method in [method for method in methods
                       if getattr(plugin_class, method) == getattr(CommandPlugin, method)]:
            logger.error("Plugin {} does not override {} ".format(plugin_class.__name__, method))
            return False
        return True

    def _iter_plugin_entry_points(self):
        from pkg_resources import iter_entry_points
        for entry_point in iter_entry_points(COMMAND_PLUGIN_ENTRY_POINT):
            try:
                yield entry_point.load()
            except Exception:
                logger.error("There was a problem loading plugin from entry point {!r}".format(entry_point))

    def get_plugin_clases_from_entry_points(self):
        return list(self._iter_plugin_entry_points())

    def get_all_plugin_classes(self):
        return self.get_plugin_clases_from_entry_points()

plugin_repository = PluginRepository()
